
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/poll.h>
#include <fcntl.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>

#include "../../src/initng_global.h"
#include "../../src/initng_process_db.h"
#include "../../src/initng_service_cache.h"
#include "../../src/initng_handler.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/initng_active_db.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_load_module.h"
#include "../../src/initng_plugin_callers.h"
#include "../../src/initng_global.h"
#include "../../src/initng_error.h"

#include "initng_stcmd.h"
#include "print_service.h"


static int cmd_get_pid_of(char *arg);
static int cmd_start(char *arg);
static int cmd_start_on_new(char *arg);
static int cmd_free_service(char *arg);
static int cmd_stop(char *arg);
static int cmd_restart(char *arg);
static void cmd_print_uptime(char *arg, FILE * fd);
static int cmd_reload(char *arg);
static int cmd_initng_restart(char *arg);
static int cmd_initng_quit(char *arg);
static void cmd_print_modules(char *arg, FILE * fd);
static int cmd_load_module(char *arg);
static int cmd_unload_module(char *arg);
static void cmd_print_service_db(char *arg, FILE * fd);
static void cmd_print_active_db(char *arg, FILE * fd);
static int cmd_percent_done(char *arg);
static int cmd_set_pid(char *arg);
static int cmd_stop_unneeded(char *arg);
static void cmd_get_father_of(char *arg, FILE * fd);
static void cmd_get_depends_on(char *arg, FILE * fd);
static void cmd_get_depends_off(char *arg, FILE * fd);

#ifdef DEBUG
static int cmd_toggle_verbose(char *arg);
static int cmd_add_verbose(char *arg);
static int cmd_del_verbose(char *arg);
#endif


s_command GET_PID_OF =
    { 'g', "get_pid_of", INT_COMMAND, REQUIRES_OPT,
{(void *) &cmd_get_pid_of}, "Get pid of service" };
s_command START_SERVICE =
    { 'u', "start", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_start}, "Start service" };
s_command START_ON_NEW =
    { 'j', "restart_from", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_start_on_new},
"Stop all services, and start from"
};
s_command FREE_SERVICE =
    { 'z', "zap", TRUE_OR_FALSE_COMMAND, USES_OPT,
{(void *) &cmd_free_service},
"Free service"
};
s_command STOP_SERVICE =
    { 'd', "stop", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT, {(void *) &cmd_stop},
    "Stop service" };
s_command RESTART_SERVICE =
    { 'r', "restart", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_restart},
"Restart service"
};

s_command PRINT_UPTIME =
    { 't', "time", PRINT_COMMAND, REQUIRES_OPT, {(void *) &cmd_print_uptime},
    "Print uptime" };
s_command SERVICE_RELOAD =
    { 'R', "reload_service", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_reload}, "Service reload" };
s_command INITNG_RESTART =
    { 'x', "initng_restart", TRUE_OR_FALSE_COMMAND, NO_OPT,
{(void *) &cmd_initng_restart},
"Restarts initng"
};
s_command QUIT_INITNG =
    { 'q', "quit", TRUE_OR_FALSE_COMMAND, NO_OPT, {(void *) &cmd_initng_quit},
"Quits initng"
};
s_command PRINT_MODULES =
    { 'm', "print_modules", PRINT_COMMAND, NO_OPT,
{(void *) &cmd_print_modules},
"Print loaded modules"
};
s_command LOAD_MODULE =
    { 'o', "load_module", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_load_module},
"Load Module"
};
s_command UNLOAD_MODULE =
    { 'w', "unload_module", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_unload_module},
"UnLoad Module"
};
s_command PRINT_SERVICE_DB =
    { 'p', "print_service_db", PRINT_COMMAND, NO_OPT,
{(void *) &cmd_print_service_db},
"Print service_db"
};

s_command PRINT_ACTIVE_DB =
    { 'A', "print_active_db", PRINT_COMMAND, NO_OPT,
{(void *) &cmd_print_active_db},
"Print active_db"
};

s_command PERCENT_DONE =
    { 'n', "done", INT_COMMAND, NO_OPT, {(void *) &cmd_percent_done},
"Prints percent of system up"
};
s_command SET_PID =
    { 'P', "set_pid", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_set_pid},
"Set pid to a service"
};
s_command STOP_UNNEEDED =
    { 'y', "stop_unneeded", TRUE_OR_FALSE_COMMAND, NO_OPT,
{(void *) &cmd_stop_unneeded},
"Stop all services, not in runlevel"
};

s_command GET_FATHER_OF =
    { 'f', "father", PRINT_COMMAND, NO_OPT, {(void *) &cmd_get_father_of},
    "Print father to" };

s_command DEPENDS_ON =
    { 'a', "service_dep_on", PRINT_COMMAND, REQUIRES_OPT,
{(void *) &cmd_get_depends_on},
"Print what services me depends on"
};

s_command DEPENDS_OFF =
    { 'b', "service_dep_on_me", PRINT_COMMAND, REQUIRES_OPT,
{(void *) &cmd_get_depends_off},
"Print what dependencies that are depending on me"
};


#ifdef DEBUG
s_command TOGGLE_VERBOSE =
    { 'v', "verbose", TRUE_OR_FALSE_COMMAND, NO_OPT,
{(void *) &cmd_toggle_verbose},
"Toggle the verbose flag"
};
s_command ADD_VERBOSE =
    { 'i', "add_verbose", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_add_verbose},
"Add string to watch for to make initng verbose"
};
s_command DEL_VERBOSE =
    { 'k', "del_verbose", TRUE_OR_FALSE_COMMAND, REQUIRES_OPT,
{(void *) &cmd_del_verbose},
"Del string to watch for to make initng verbose"
};
#endif


static int cmd_get_pid_of(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (0);

    if ((apt = active_db_find_in_name(arg)) &&
        apt && apt->start_process && apt->start_process->pid)
        return (apt->start_process->pid);
    return (0);
}


static int cmd_start(char *arg)
{
    if (!arg)
        return (FALSE);

    return (start_new_service_named(arg));
}


static int cmd_start_on_new(char *arg)
{
    if (!arg)
        return (FALSE);

    g.when_out = THEN_NEW_RUNLEVEL;
    g.runlevel = i_strdup(arg);
    stop_all();
    return (TRUE);
}

static int cmd_free_service(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);
    apt = active_db_find_in_name(arg);
    if (!apt)
        return (FALSE);

    active_db_del(apt);
    return (TRUE);
}

static int cmd_stop(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);
    apt = active_db_find_in_name(arg);
    if (!apt)
        return (FALSE);

    return (stop_service(apt));
}

static int cmd_restart(char *arg)
{
    active_h *apt = NULL;

    if (!arg)
        return (FALSE);
    apt = active_db_find_in_name(arg);
    if (!apt)
        return (FALSE);

    return (restart_service(apt));
}





static void cmd_print_uptime(char *arg, FILE * fd)
{
    active_h *apt = NULL;

    if (!arg)
        return;

    apt = active_db_find_in_name(arg);
    if (!apt)
        return;

    fprintf(fd, "Uptime is %f sec.\n",
            difftime(time(0), apt->time_got_status));
    return;
}


static int cmd_reload(char *arg)
{
    service_h *s;

    s = service_db_find_in_name(arg);
    if (!s)
    {
        D_("Clearing complete db.\n");
        service_db_free_all();
        return (TRUE);
    }

    D_("removing service data for %s!\n", arg);
    service_db_free(s);
    return (TRUE);
}

static int cmd_initng_restart(char *arg)
{
    (void) arg;
    g.when_out = THEN_RESTART;
    stop_all();
    return (TRUE);
}

static int cmd_initng_quit(char *arg)
{
    (void) arg;
    g.when_out = THEN_QUIT;
    stop_all();
    return (TRUE);
}



static void cmd_print_modules(char *arg, FILE * fd)
{
    m_h *mod = NULL;

    (void) arg;

    fprintf(fd, "modules: \n");

    while ((mod = initng_get_next_module(mod)))
        fprintf(fd, "%s\n", mod->module_name);

    return;
}

static int cmd_load_module(char *arg)
{
    if (!arg)
        return (FALSE);

    return (initng_load_module(arg));
}

static int cmd_unload_module(char *arg)
{
    if (!arg)
        return (FALSE);

    return (initng_unload_module_named(arg));
}

static void cmd_print_service_db(char *arg, FILE * fd)
{
    service_h *s;

    if (arg && (s = service_db_find_in_name(arg)))
    {
        service_db_print(s, fd);
    }
    else if (arg)
    {
        fprintf(fd, "No such service: %s\n", arg);
    }
    else
    {
        service_db_print_all(fd);
    }

    return;
}

static void cmd_print_active_db(char *arg, FILE * fd)
{
    active_h *s;

    if (arg && (s = active_db_find_in_name(arg)))
    {
        active_db_print(s, fd);
    }
    else if (arg)
    {
        fprintf(fd, "No such service: %s\n", arg);
    }
    else
    {
        active_db_print_all(fd);
    }

    return;
}

static int cmd_percent_done(char *arg)
{
    (void) arg;

    return (active_db_percent_started());
}

static int cmd_set_pid(char *arg)
{
    active_h *apt = NULL;
    char *sname = NULL;
    int pid = 0;
    sname = initng_calloc(100, sizeof(char));
    sscanf(arg, "%i:%s", &pid, sname);
    if (!pid || !sname)
        return (FALSE);
    apt = active_db_find_in_name(sname);
    if (apt && apt->name && apt->start_process)
    {
        D_("setting pid of %s from %i to %i!\n",
           apt->name, apt->start_process->pid, pid);
        apt->start_process->pid = pid;
        if (active_db_is(&UP_WHEN_PID_SET, apt))
            apt->a_status = RUNNING;
        return (TRUE);
    }


    F_("did not found service %s to set pid %i to.", sname, pid);
    return (FALSE);
}


static int cmd_stop_unneeded(char *arg)
{
    (void) arg;
    stop_unneeded_services();
    return (TRUE);
}

static void cmd_get_father_of(char *arg, FILE * fd)
{
    service_h *s;

    s = service_db_find_in_name(arg);
    if (!s)
    {
        F_("Did not find service: %s\n", arg);
        fprintf(fd, "Can't find service: %s\n", arg);
        return;
    }

    if (s->father_name)
        fprintf(fd, "Father to service %s is %s\n", s->name, s->father_name);
    else
        fprintf(fd, "Service %s has no father\n", s->name);

    return;
}


static void cmd_get_depends_on(char *arg, FILE * fd)
{
    active_h *a = NULL;
    active_h *dep_s = active_db_find_in_name(arg);

    if (dep_s)
    {
        fprintf(fd, "The \"%s\" depends on:\n", dep_s->name);
        while ((a = active_db_next_service_this_depends_on(dep_s, a)))
        {
            fprintf(fd, "  %s\t\t[%s]\n", a->name,
                    active_db_get_status_string(a->a_status));
        }
        fprintf(fd, "\n\n");
        return;
    }
    fprintf(fd, "Did not find service!\n");
    return;
}

static void cmd_get_depends_off(char *arg, FILE * fd)
{
    active_h *a = NULL;
    active_h *dep_s = active_db_find_in_name(arg);

    if (dep_s)
    {
        fprintf(fd, "The \"%s\" depends on:\n", dep_s->name);
        fprintf(fd, "Services that is depending on \"%s\" is:\n",
                dep_s->name);
        while ((a = active_db_next_dep_on_me(dep_s, a)))
        {
            fprintf(fd, "  %s\t\t[%s]\n", a->name,
                    active_db_get_status_string(a->a_status));
        }
        fprintf(fd, "\n\n");
        return;
    }
    fprintf(fd, "Did not find service!\n");
    return;


}

#ifdef DEBUG


static int cmd_toggle_verbose(char *arg)
{
    (void) arg;
    switch (g.verbose)
    {
        case 0:
            g.verbose = 1;
            break;
        case 1:
            g.verbose = 0;
            break;
        case 2:
            g.verbose = 3;
            break;
        case 3:
            g.verbose = 2;
            break;
        default:
            g.verbose = 0;
            W_("Unknow verbose id %i\n", g.verbose);
            break;
    }
    return (g.verbose);
}

static int cmd_add_verbose(char *arg)
{
    if (!arg)
        return (FALSE);

    return (verbose_add(arg));
}

static int cmd_del_verbose(char *arg)
{
    if (!arg)
        return (FALSE);

    return (verbose_del(arg));
}
#endif

int module_init(const char *version)
{
    D_("module_init(stcmd);\n");
    if (strcmp(version, INITNG_VERSION) != 0)
    {
        F_("This module, is compiled for \"%s\" version, and initng is compiled on \"%s\" version, wont load this module!\n", INITNG_VERSION, version);
        return (FALSE);
    }

    initng_command_add(&GET_PID_OF);
    initng_command_add(&START_SERVICE);
    initng_command_add(&START_ON_NEW);
    initng_command_add(&FREE_SERVICE);
    initng_command_add(&STOP_SERVICE);
    initng_command_add(&RESTART_SERVICE);
    initng_command_add(&PRINT_UPTIME);
    initng_command_add(&SERVICE_RELOAD);
    initng_command_add(&INITNG_RESTART);
    initng_command_add(&QUIT_INITNG);
    initng_command_add(&PRINT_MODULES);
    initng_command_add(&LOAD_MODULE);
    initng_command_add(&UNLOAD_MODULE);
    initng_command_add(&PRINT_SERVICE_DB);
    initng_command_add(&PRINT_ACTIVE_DB);
    initng_command_add(&PERCENT_DONE);
    initng_command_add(&SET_PID);
    initng_command_add(&STOP_UNNEEDED);
    initng_command_add(&GET_FATHER_OF);
    initng_command_add(&DEPENDS_ON);
    initng_command_add(&DEPENDS_OFF);
#ifdef DEBUG
    initng_command_add(&TOGGLE_VERBOSE);
    initng_command_add(&ADD_VERBOSE);
    initng_command_add(&DEL_VERBOSE);
#endif


    D_("libstcmd.so.0.0 loaded!\n");
    return (TRUE);
}


void module_unload(void)
{
    D_("module_unload(stcmd);\n");

    initng_command_del(&GET_PID_OF);
    initng_command_del(&START_SERVICE);
    initng_command_del(&START_ON_NEW);
    initng_command_del(&FREE_SERVICE);
    initng_command_del(&STOP_SERVICE);
    initng_command_del(&RESTART_SERVICE);
    initng_command_del(&PRINT_UPTIME);
    initng_command_del(&SERVICE_RELOAD);
    initng_command_del(&INITNG_RESTART);
    initng_command_del(&QUIT_INITNG);
    initng_command_del(&PRINT_MODULES);
    initng_command_del(&LOAD_MODULE);
    initng_command_del(&UNLOAD_MODULE);
    initng_command_del(&PRINT_SERVICE_DB);
    initng_command_del(&PRINT_ACTIVE_DB);
    initng_command_del(&PERCENT_DONE);
    initng_command_del(&SET_PID);
    initng_command_del(&STOP_UNNEEDED);
    initng_command_del(&GET_FATHER_OF);
    initng_command_del(&DEPENDS_ON);
    initng_command_del(&DEPENDS_OFF);
#ifdef DEBUG
    initng_command_del(&TOGGLE_VERBOSE);
    initng_command_del(&ADD_VERBOSE);
    initng_command_del(&DEL_VERBOSE);
#endif


    D_("libstcmd.so.0.0 unloaded!\n");

}
