
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <sys/types.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <alloca.h>
#include <assert.h>
#include "initng_active_db.h"
#include "initng_service_cache.h"
#include "initng_process_db.h"
#include "initng_global.h"
#include "initng_toolbox.h"
#include "initng_common.h"
#include "initng_static_data_id.h"



void actictive_db_change_active_h(active_h * from, active_h * to);
void actictive_db_change_service_h(service_h * from, service_h * to);


/* #####  ACTIVE_DB_FIND_BY   ######## */

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_name(const char *service)
{
    active_h *current = g.active_database;

    assert(service);

    /* first, search for an exact match */
    while (current)
    {
        /* check if this service name is like service */
        if (strcmp(current->name, service) == 0)
            return (current);               /* return it */

        /* jump to next service if any */
        current = current->next;
    }

    /* did not find any */
    return NULL;
}

active_h *active_db_find_in_name(const char *service)
{
    active_h *current = g.active_database;

    assert(service);

    /* first, search for an exact match */
    while (current)
    {
        /* check if this service name is like service */
        if (strcmp(current->name, service) == 0)
            return (current);               /* return it */

        /* jump to next service if any */
        current = current->next;
    }

    /* then search for a word match */
    current = g.active_database;
    while (current)
    {
        /* check if this service name is like service */
        if (strstr(current->name, service))
        {
            D_("Warning, the return is not an exact match!\n");
            return (current);               /* return it */
        }
        /* jump to next service if any */
        current = current->next;
    }


    /* did not find any */
    return NULL;
}

/* return index of service in active data structure or -1 if not found */
active_h *active_db_find_by_service_h(service_h * service)
{
    active_h *current = g.active_database;

    assert(service);

    while (current)
    {
        /* check if this service->from_service is like service */
        if (current->from_service == service)
            return (current);               /* return it */

        /* jump to next service if any */
        current = current->next;
    }

    return NULL;
}

/* returns pointer to active_h process belongs to, and sets process_type */
active_h *active_db_find_by_pid(pid_t pid)
{
    active_h *current = g.active_database;

    while (current)
    {
        /* check for matching */
        if (current->start_process && current->start_process->pid == pid)
            return (current);
        if (current->stop_process && current->stop_process->pid == pid)
            return (current);

        /* jump to next service if any */
        current = current->next;
    }

    /* bad luck */
    return (NULL);
}





/* #########   STRUCT/CHAIN HANDLE  ######## */

/* add active to data structure */
/* remember to free service, if this fails */
active_h *active_db_add(active_h * add_this)
{
    active_h *current = NULL;

    /* we have to get this data */
    assert(add_this);

    /* Look for duplicate */
    if ((current = active_db_find_by_name(add_this->name)))
    {
        W_("active_db_add(%s): duplicate here\n", add_this->name);
        return (NULL);
    }

    /* check if this is the initial */
    if (!g.active_database)
    {
        g.active_database = add_this;
        D_("service %s is added as the initial entry.\n", add_this->name);
        return (g.active_database);
    }

    /* go trhu every service, to add this on the end */
    current = g.active_database;
    while (current)
    {
        /* this may be an empty spot */
        if (current->name == NULL && current->a_status == NONE)
        {
            /* save the next pointer */
            active_h *tmp_next = current->next;

            /* copy the add_this, over to the free spot */
            memcpy(current, add_this, sizeof(active_h));

            /* set next again, to the old value */
            current->next = tmp_next;

            /* free add_this */
            free(add_this);

            /* return, new adress */
            return (current);
        }

        /* fast forward */
        if (current->next)
        {
            current = current->next;
            continue;
        }

        /* if we got here, current->next is free */
        current->next = add_this;

        /* make sure next, after this newly added is NULL. */
        current->next->next = NULL;
        D_("service %s is added after %s.\n", add_this->name, current->name);
        return (current->next);
    }

    F_("did leave while loop, this is bad, check it up!\n");
    return (NULL);
}

/* ############    ACTIVE_DB_ALLOC/FREE ##########  */

/* creates a new active_h at new_active */
active_h *active_db_new(const char *name)
{
    active_h *new_active = NULL;

    /* allocate a new active entry */
    new_active = (active_h *) initng_calloc(1, sizeof(active_h));       /* Allocate memory for a new active */

    /* out of memory? */
    if (!new_active)
    {
        F_("active_db_new(): Unable to allocate active, out of memory?\n");
        return (NULL);
    }

    new_active->name = i_strdup(name);
    new_active->data = (s_data *) initng_calloc(1, sizeof(s_data));
    new_active->data->next = NULL;
    new_active->next = NULL;

    new_active->time_got_status = 0;
    new_active->a_status = NONE;
    new_active->from_service = NULL;

    if (new_active->start_process)
        process_db_free(&new_active->start_process);
    if (new_active->stop_process)
        process_db_free(&new_active->stop_process);

    return (new_active);
}

active_h *active_db_del_real(active_h * db, active_h * service)
{
    active_h *next;

    if (db == service)
    {
        D_("(%s);\n", db->name);
        next = db->next;
        active_db_free(db);
        return next;
    }
    if (db->next)
        db->next = active_db_del_real(db->next, service);
    else
        F_("Failed to remove service %s from struct!, cant free!\n",
           service->name);
    return db;
}

/* free some values in this one */
void active_db_free(active_h * pf)
{
    assert(pf);

    D_("(%s);\n", pf->name);

    /* try free process structs, data and name */
    if (pf->start_process)
        process_db_free(&pf->start_process);
    if (pf->stop_process)
        process_db_free(&pf->stop_process);
    if (pf->data)
        d_remove_all(pf->data);
    if (pf->name)
        free(pf->name);

    /* try free->next, if its NULL */
    if (pf->next && pf->next->name == NULL && pf->next->a_status == NONE)
        active_db_free(pf->next);

    /*
     * if this is not the last entry, we cant free this one, 
     *  becouse its a part of a structure with a ->next 
     */
    if (pf->next)
    {
        /* reset active_h */
        pf->name = NULL;
        pf->from_service = NULL;
        pf->a_status = NONE;
        pf->start_process = NULL;
        pf->stop_process = NULL;
        pf->data = NULL;
        pf->m = 0;

    }
    else
    {
        /* free active_h */
        free(pf);
    }
}


/* clear database */
void active_db_free_all(void)
{
    active_h *current = g.active_database;
    active_h *next = NULL;

    while (current)
    {
        /* chek if there is a next */
        next = current->next;

        /* free this one */
        active_db_del(current);

        /* step forward */
        current = next;
    }

    return;
}

/* #########  ACTIVE_DB_UTILS  #############  */

/* compensate time */
void active_db_compensate_time(time_t skew)
{
    active_h *current = g.active_database;

    while (current)
    {
        /* change this time */
        current->time_got_status += skew;
        current = current->next;
    }
}


/* update all service_h pointers in whole active_db */
void active_db_change_service_h(service_h * from, service_h * to)
{
    active_h *current = g.active_database;

    while (current)
    {
        /* change this time */
        if (current->from_service == from)
            current->from_service = to;

        current = current->next;
    }
}


/* Walks thru db, and returns all services, that depends on this service */
active_h *active_db_next_dep_on_me(active_h * depend, active_h * current)
{
    char *string = NULL;

    assert(depend);
    assert(depend->name);

    /* set current if not set */
    if (!current)
        current = g.active_database;
    /* or youst take the next one */
    else
        current = current->next;

    while (current)
    {
        /* if its me, jump to next one */
        if (current == depend)
        {
            current = current->next;
            continue;
        }

        string = NULL;
        /* check if this service has any NEED flags */
        while ((string = active_db_get_next_string(&NEED, current, string)))
        {
            if (strcmp(depend->name, string) == 0)
            {
                D_("service %s depends on me!\n", depend->name);
                return (current);
            }
        }

        string = NULL;
        /* check if this service has any USE flags */
        while ((string = active_db_get_next_string(&USE, current, string)))
        {
            assert(string);
            if (strcmp(depend->name, string) == 0)
            {
                D_("service %s depends on me!\n", depend->name);
                return (current);
            }
        }

        /* jump to next one */
        current = current->next;

    }

    /* return negative */
    return (NULL);
}



/* return next active, that is in use or need flags on depend service */
active_h *active_db_next_service_this_depends_on(active_h * depend,
                                                 active_h * current)
{
    char *string = NULL;
    active_h *this_active = NULL;

    assert(depend);

    while ((string = active_db_get_next_string(&NEED, depend, string)))
    {
        if ((this_active = active_db_find_by_name(string)))
        {
            if (this_active == current)
            {
                current = NULL;
                continue;
            }
            /* we have to reach current, before returning anything */
            if (current)
                continue;

            D_("service: %s need %s found %s\n", depend->name, string,
               this_active->name);
            return (this_active);
        }
    }

    string = NULL;
    while ((string = active_db_get_next_string(&USE, depend, string)))
    {
        if ((this_active = active_db_find_by_name(string)))
        {
            if (this_active == current)
            {
                current = NULL;
                continue;
            }
            /* we have to reach current, before returning anything */
            if (current)
                continue;

            D_("service: %s use %s found %s\n", depend->name, string,
               this_active->name);
            return (this_active);
        }
    }

    return (NULL);
}


/* active_db_count counts a type, if null, count all */
int active_db_count(e_a_status a_status_to_count)
{
    int counter = 0;            /* actives counter */
    active_h *current = g.active_database;

    if (a_status_to_count == NONE)
    {
        /* ok, go thru all */
        while (current)
        {

            /* count almost all */
            switch (current->a_status)
            {
                case FAIL_STARTING:
                case FAIL_STOPPING:
                case START_DEP_FAILED:
                case STOP_DEP_FAILED:
                case STOPPED:
                    break;
                default:
                    counter++;
                    break;
            }

            /* go to next */
            current = current->next;
        }
    }
    else
    {
        /* ok, go thru all */
        while (current)
        {
            /* check if this is the status to count */
            if (current->a_status == a_status_to_count)
                counter++;

            /* go to next */
            current = current->next;
        }
    }

    /* return counter */
    return (counter);
}


/* calculate percent of processes started */
int active_db_percent_started(void)
{
    int starting = 0;
    int up = 0;
    float tmp = 0;

    active_h *current = g.active_database;

    while (current)
    {
        switch (current->a_status)
        {
            case STARTING:
            case WAITING_FOR_START_DEP:
            case START_DEP_MET:
            case LOADING:
                D_("active_db_percent_started(): %s starting\n",
                   current->name);
                starting++;
                break;
            case DONE:
            case RUNNING:
                D_("active_db_percent_started(): %s: up\n", current->name);
                up++;
                break;
            default:
                break;
        }
        current = current->next;
    }
    D_("active_db_percent_started(): up: %i   starting: %i\n", up, starting);

    /* if no one starting */
    if (starting <= 0)
        return (100);

    if (up > 0)
    {
        tmp = 100 * (float) up / (float) (starting + up);
        D_("active_db_percent_started(): up/starting: %f percent: %i\n\n",
           (float) up / (float) starting, (int) tmp);
        return ((int) tmp);
    }
    return 0;
}

/* calculate percent of processes stopped */
int active_db_percent_stopped(void)
{
    int stopping = 0;
    int down = 0;
    float tmp = 0;
    active_h *current = g.active_database;

    while (current)
    {
        switch (current->a_status)
        {
            case STOPPING:
            case WAITING_FOR_STOP_DEP:
            case STOP_DEP_MET:
            case STOP_MARKED:
                D_("active_db_percent_stopped(): %s stopping\n",
                   current->name);
                stopping++;
                break;
            case STOPPED:
                D_("active_db_percent_stopped(): %s: down\n", current->name);
                down++;
                break;
            default:
                break;
        }
        current = current->next;
    }

    D_("active_db_percent_stopped(): down: %i   stopping: %i\n", down,
       stopping);

    /* if no one stopping */
    if (stopping <= 0)
        return (100);

    if (down > 0)
    {
        tmp = 100 * (float) down / (float) (stopping + down);
        D_("active_db_percent_stopped(): down/stopping: %f percent: %i\n\n",
           (float) down / (float) stopping, (int) tmp);
        return ((int) tmp);
    }
    return 0;
}


/* ########     Data handlers  ############## */
char *active_db_get_string(s_entry * type, active_h * from_active)
{
    char *t = NULL;

    assert(from_active);
    assert(from_active->data);

    /* look in active datas */
    if ((t = d_get_string(type, from_active->data)))
        return (t);

    /* load a service file, if not exists */
    if (!from_active->from_service)
        get_service(from_active);

    /* look in service file data */
    if (from_active->from_service)
        return (service_db_get_string(type, from_active->from_service));

    return (NULL);
}

int active_db_get_int(s_entry * type, active_h * from_active)
{
    assert(from_active);
    assert(from_active->data);

    if (d_is(type, from_active->data))
        return (d_get_int(type, from_active->data));

    if (!from_active->from_service)
        get_service(from_active);
    if (from_active->from_service)
        return (service_db_get_int(type, (from_active->from_service)));
    return (0);
}

char *active_db_get_next_string(s_entry * type, active_h * from_active,
                                char *last)
{
    char *r = NULL;

    assert(from_active);
    assert(from_active->data);

    /* check active overlay data */
    if ((r = d_get_next_string(type, from_active->data, last)))
        return (r);

    /* load service from disk, if nonexistant */
    if (!from_active->from_service)
        get_service(from_active);

    /* check service data */
    if (from_active->from_service)
        return (service_db_get_next_string
                (type, from_active->from_service, last));
    return (NULL);
}


int active_db_is(s_entry * type, active_h * from_active)
{
    assert(from_active);
    assert(from_active->data);

    /* check active overlay data */
    if (d_is(type, from_active->data))
        return (TRUE);

    /* load service from disk, if nonexistant */
    if (!from_active->from_service)
        get_service(from_active);

    /* check service data */
    if (from_active->from_service)
        if (service_db_is(type, from_active->from_service))
            return (TRUE);

    return (FALSE);
}

int active_db_count_type(s_entry * type, active_h * from_active)
{
    assert(from_active);
    if (!from_active->from_service)
        get_service(from_active);
    if (from_active->from_service)
        return (service_db_count_type(type, from_active->from_service));
    return (0);
}

void active_db_set_string(s_entry * type, active_h * from_active,
                          char *string)
{
    assert(from_active);
    d_set_string(type, from_active->data, string);
}

void active_db_set(s_entry * type, active_h * from_active)
{
    assert(from_active);
    d_set(type, from_active->data);
}

void active_db_remove(s_entry * type, active_h * from_active)
{
    assert(from_active);
    W_("TODO active_db_remove(%s, %s);\n", type->opt_name, from_active->name);
}



 /* ##############  MARK SERVICE  ################ */
void active_db_unmark_all(int mark_type)
{
    active_h *current = g.active_database;

    while (current)
    {
        if (current->m / mark_type)
            current->m -= mark_type;
    }
}

void active_db_mark(active_h * service, int mark_type)
{
    assert(service);
    if ((service->m) / mark_type)
        return;

    service->m += mark_type;
}

void active_db_unmark(active_h * service, int mark_type)
{
    assert(service);
    if (!((service->m) / mark_type))
        return;

    service->m -= mark_type;
}
