
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* common standard first define */
#include "initng.h"

/* system headers */
#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

/* local headers include */
#include "initng_main.h"
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_common.h"
#include "initng_toolbox.h"
#include "initng_string_tools.h"


/*
 * this fuction walks thru the g.Argv, if it founds service name,
 * with an starting -service, this function will return FALSE, and
 * the service wont be able to load, meens that is it blacklisted.
 * return TRUE if service is okay to load.
 */
int service_blacklisted(const char *name)
{
    int i;

    assert(name);
    assert(g.Argv);
    assert(*g.Argv);

    /* walk through arguments looking for this dep to be blacklisted */
    for (i = 1; (g.Argv)[i]; i++)
    {
        /* if we got a match */
        if ((g.Argv)[i][0] == '-' && strstr(name, (g.Argv)[i] + 1))
        {
            return (TRUE);
        }
    }
    return (FALSE);
}

/* creates a active service */
active_h *load_to_active(const char *service_name)
{
    active_h *a_new = NULL;
    active_h *a_new_p = NULL;

    assert(service_name);
    D_("load_to_active(%s); \n", service_name);

    /* check the blacklist db */
    if (service_blacklisted(service_name))
    {
        F_("load_to_active(%s): Service BLACKLISTED.\n", service_name);
        return (NULL);
    }

    /* Make sure there isn't any process with this name running */
    if ((a_new = active_db_find_by_name(service_name)))
    {
        D_("load_to_active(%s): process exists, returning pointer to it. \n",
           service_name);
        return (a_new);
    }

    /* if not create a new active entry, out of memory? */
    if (!(a_new = active_db_new(service_name)))
    {
        D_("load_to_active(%s): Unable to allocate process, out of memory?\n",
           service_name);
        return (NULL);
    }

    /* this actually loads service from disk */
    if (!get_service(a_new) || !(a_new->from_service))
    {
        F_("load_to_active(%s): Can't get service!\n", service_name);
        if (a_new)
            active_db_free(a_new);
        return (NULL);
    }

    if (a_new->from_service->type != TYPE_SERVICE &&
        a_new->from_service->type != TYPE_RUNLEVEL)
    {
        F_("load_to_active(%s): service is not a service type!\n",
           service_name);
        if (a_new)
            active_db_free(a_new);
        return (NULL);
    }


    /* return with pointer to service */
    if ((a_new_p = active_db_add(a_new)))
        return (a_new_p);

    /* fail by defaukt */
    if (a_new)
        active_db_free(a_new);
    return (NULL);
}



/* this function tries to find the from_service pointer from a service name.
   this actually loads the config from disk, parser by parser */
int get_service(active_h * service)
{
    int i = 0;

    assert(service);
    assert(service->name);
    D_("get_service(%s);\n", service->name);

    /* check if service is set already */
    if (service->from_service)
    {
        D_("get_service(%s): already have a service!\n", service->name);
        return (TRUE);
    }

    /* check cache database */
    if ((service->from_service = service_db_find_by_name(service->name)))
    {
        D_("get_service(%s): found service in service database!\n",
           service->name);
        return (TRUE);
    }

    /* try parse with dynamic loaded parsers */
    if (!(service->from_service = initng_modules_parse(service->name)))
    {
        /* or fail */
        D_("get_service(%s): Can't get source.. \n", service->name);
        return (FALSE);
    }

    D_(" get_service(%s): got service from initng dynamic parsers.\n",
       service->name);

    /* will continue here only, if initng_modules_parse did suceed */

    /*
     * if service->name is "samba", and service->from_service->name is "daemon/samba" 
     * we have to update service->name.
     * also notice that service->name shud not exactly match service->from_service->name, 
     * service->name might be "eth0" and service->from_service->name might be "net / *".
     */

    /* cont the tokens to the first '/' */
    while (service->from_service->name[i]
           && service->from_service->name[i] != '/')
        i++;
    /* if there was no '/' dont complain */
    if (service->from_service->name[i] != '/')
        return (TRUE);

    /* compare the first i chars */
    if (strncmp(service->from_service->name, service->name, i) != 0)
    {
        D_("Name %s ..\n", service->name);
        char *new_name = (char *) initng_calloc(strlen(service->name) + i + 3,
                                                sizeof(char));

        strncpy(new_name, service->from_service->name, i + 1);
        strcat(new_name, service->name);
        free(service->name);
        service->name = new_name;
        D_(" Is now: %s \n", service->name);
    }

    /* return happily */
    return (TRUE);

}


/*
 * when a service fails to start, this function
 * walks thru all its dependencys, and mark it "dependency failed to start"
 */
static void dep_failed_to_start(active_h * service)
{
    active_h *next = NULL;

    while ((next = active_db_next_dep_on_me(service, next)))
    {
        mark_service(next, START_DEP_FAILED);
        dep_failed_to_start(next);
    }
}

/*
 * when a service fails to stop, this function
 * walks thru all its dependencys, and mark it "dependency failed to stop"
 */
static void dep_failed_to_stop(active_h * service)
{
    active_h *next = NULL;

    while ((next = active_db_next_dep_on_me(service, next)))
    {
        mark_service(next, STOP_DEP_FAILED);
        dep_failed_to_start(next);
    }
}

/*
 * Use this function to change the status of an service, this
 * function might refuse to change to that state, and if that
 * it will return FALSE, please always check the return value
 * when calling this funcion.
 */
int mark_service(active_h * service, e_a_status status)
{
    time_t old_time;

    assert(service);
    /* if already set */
    if (service->a_status == status)
    {
        D_("warning, this status %i is already set on %s!\n", status,
           service->name);
        /* it is actually already set to requested status */
        return (TRUE);
    }


    D_("mark_service(%s, %s);\n", service->name,
       active_db_get_status_string(status));

    /* set status and time */
    service->a_status = status;
    old_time = service->time_got_status;
    service->time_got_status = time(0);

    /* call all dynamic modules that want to know about this */
    if (!initng_load_module_service_changed(service))
    {
        W_("Some plugin did fail!\n");
    }

    /* is it really changed? */
    if (service->a_status != status)
    {
        D_("Some plugin did not want this to change!\n");
        service->time_got_status = old_time;
        return (FALSE);
    }



    /* if g.interrupt is set, main loop will try
       start new services */
    switch (service->a_status)
    {
        case DONE:
        case RUNNING:
            /* check if this is the final service */
            if (strcmp(service->name, g.runlevel) == 0)
                set_sys_state(STATE_UP);
            /* set interrupt, so main loop will check other
               that waits for deps */
            g.interrupt = TRUE;
            break;
        case STOPPED:
            /* set interrupt, so main loop will check other
               that waits for deps */
            g.interrupt = TRUE;
            break;
        case FAIL_STARTING:
            dep_failed_to_start(service);
            g.interrupt = TRUE;
            break;
        case FAIL_STOPPING:
            dep_failed_to_stop(service);
            g.interrupt = TRUE;
            break;
        case STOP_DEP_FAILED:
        case START_DEP_FAILED:
            g.interrupt = TRUE;
            break;
        default:
            break;
    }

    /* add to history - Only if its actually changed */
    if (service->a_status != status)
    {
        D_("Did not change.\n");
        return (FALSE);
    }

    history_add_values(service);

    D_("service %s is now %s.\n", service->name,
       active_db_get_status_string(service->a_status));
    return (TRUE);
}


/*
 * This function will change variables in from, like $NAME, $SERVICE, $CLASS
 * into its right content, and return a new allocated string with the result.
 */
char *fix_variables(char *from, active_h * s)
{
    int i = 0;
    int j = 0;
    char *tmp;
    char *to;

    if (!from)
        return (NULL);
    assert(s);
    assert(s->name);
    to = (char *) initng_calloc((strlen(from) + 150), sizeof(char));
    while (from[i] != '\0')
    {
        if (from[i] == '$')
        {
            i++;
            if (strncmp(&from[i], "NAME", strlen("NAME")) == 0)
            {
                tmp = st_strip_path(s->name);
                strcpy(&to[j], tmp);
                i = i + strlen("NAME");
                j = j + strlen(tmp);
            }
            else if (strncmp(&from[i], "SERVICE", strlen("SERVICE")) == 0)
            {
                strcpy(&to[j], s->name);
                i = i + strlen("SERVICE");
                j = j + strlen(s->name);
            }
            else if (strncmp(&from[i], "CLASS", strlen("CLASS")) == 0)
            {
                tmp = st_get_path(s->name);
                strcpy(&to[j], tmp);
                i = i + strlen("CLASS");
                j = j + strlen(tmp);
                if (tmp)
                    free(tmp);
            }
            else
            {
                to[j] = from[i];
                j++;
                i++;
            }

        }
        else
        {
            to[j] = from[i];
            j++;
            i++;
        }
    }
    D_("fix_variables(): from \"%s\" to \"%s\".\n", from, to);
    return (to);
}
