/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>                          /* fcntl() */
#include "initng.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"
#include "initng_signal.h"

#include "initng_plugin_callers.h"
#include "initng_plugin_type.h"

#include "initng_fd.h"

void initng_close_all_fd(void)
{
    s_call *current = g.call_db;

    while (current)
    {
        if (current->ptype == FDWATCHERS && (current->c.fdh.fds > 0))
        {
            close(current->c.fdh.fds);
            current->c.fdh.fds = 0;
            current->c.fdh.call_module = NULL;
            current->ptype = NOPLUG;
            D_("Todo, call_del()\n");
        }
        current = current->next;
    }
}

/* Calls plugins when data is ready in pipe */
void initng_plugin_readpipe(active_h * service, int datalen,
                            process_h * process, int force_flush)
{
    s_call *current = g.call_db;

    while (current)
    {
        if (current->ptype == PIPEWATCHERS)
        {
            (*current->c.pipewatcher) (service, datalen, process,
                                       force_flush);
        }
        current = current->next;
    }
    return;
}

/* This does the actual read from pipe, all checks have been done here */
int initng_process_readpipe_read(active_h * service, process_h * x,
                                 int program_end)
{
    int forceflush = program_end;
    int maxread = MAX_BUFFER - x->buffer_pos;
    int len;

    do
    {
        errno = 0;
        len = read(x->stdout[0], &x->buffer[x->buffer_pos], maxread);
    }
    while (errno == EINTR);

    if (len > 0)
    {
        x->buffer[x->buffer_pos + len] = '\0';

        if (x->buffer_pos + len >= MAX_BUFFER)
            forceflush = TRUE;

        initng_plugin_readpipe(service, len, x, forceflush);
        x->buffer_pos = x->buffer_pos + len;
        if (forceflush)
            x->buffer_pos = 0;
    }
    return len;
}

/* This does all the fancy pancy around the read */
void initng_process_readpipe(active_h * service, process_h * x,
                             int program_end)
{
    int i;

    /* We've already quit */
    if ((!x->buffer))
        return;


    if (program_end)
    {
        int fd_flags;

        fd_flags = fcntl(x->stdout[0], F_GETFL, 0);
        fcntl(x->stdout[0], F_SETFL, fd_flags | O_NONBLOCK);
    }

    i = initng_process_readpipe_read(service, x, program_end);

    if (!program_end)
    {
        /* We have an error other than EAGAIN, set program_end */
        if ((i == -1) && (errno != EAGAIN))
        {
            program_end = TRUE;
            F_("Error reading from pipe for %s (%s)\n", service->name,
               strerror(errno));
        }
        /* End of file */
        if (i == 0)
            program_end = TRUE;
    }

    /* If i < 1 we either have 0 (end of file) or -1 (error) */
    if (program_end)
    {
        /* Program is ending, close pipes */
        initng_plugin_readpipe(service, 0, x, TRUE);
        D_("Freeing buffer for %s\n", service->name);
        if (x->stdout[0] > 0)
            close(x->stdout[0]);
        if (x->stdout[1] > 0)
            close(x->stdout[1]);
        if (x->buffer)
            free(x->buffer);
        x->stdout[0] = -1;
        x->stdout[1] = -1;
        x->buffer = NULL;
    }
}

/* this will poll all filedescriptors in g.fds[] db, for input, 
   and call the right function, when events occur 
   returns FALSE if select() timed out. */
int initng_plugin_fd_poll(int timeout)
{
    fd_set readset;
    struct timeval tv;
    int retval;
    int set = FALSE;
    active_h *tmp = g.active_database;
    s_call *current = g.call_db;

    FD_ZERO(&readset);

    /* First scan through active plugins that have listening file descriptors, and add them */
    while (current)
    {
        if (current->ptype == FDWATCHERS
            && (current->c.fdh.fds > 0 || current->c.fdh.call_module))
        {
            FD_SET(current->c.fdh.fds, &readset);
            set = TRUE;
            D_("adding fd #%i\n", current->c.fdh.fds);
        }
        current = current->next;
    }

    /* Then add all file descriptors from process read pipes */
    while (tmp)
    {
        if ((tmp->start_process) && (tmp->start_process->stdout[0] > 0))
        {
            FD_SET(tmp->start_process->stdout[0], &readset);
            set = TRUE;
        }
        if ((tmp->stop_process) && (tmp->stop_process->stdout[0] > 0))
        {
            FD_SET(tmp->stop_process->stdout[0], &readset);
            set = TRUE;
        }
        tmp = tmp->next;
    }

    /* check if there are any set */
    if (!set)
    {
        D_("No filedescriptors\n");
        sleep(timeout);
        return (TRUE);
    }

    tv.tv_sec = timeout;
    tv.tv_usec = 0;

    /* If interrupt then return and do main loop, it doesn't hurt if the filebuffers have a bit more data anyway */
    if (g.interrupt)
        return (TRUE);

    retval = select(256, &readset, NULL, NULL, &tv);
    if (g.interrupt)
        return (TRUE);
    if (retval == 0)
        return (FALSE);
    if (retval == -1)
        F_("Select returned with %d %s\n", retval, strerror(errno));

    D_("select retval=%i\n", retval);

    /* find matches, for calling right funtion to 
       fetch the incoming data */
    if (retval > 0)
    {
        D_("%d fd's active\n", retval);
        /* If a fsck is running select will always return one filehandler active, we give it 0.1 seconds to get some more data into the buffer. */
        sleep(0.1);


        /* Now scan throught callers */
        current = g.call_db;
        while (current)
        {
            if ((current->ptype == FDWATCHERS
                 && (current->c.fdh.fds > 0 || current->c.fdh.call_module))
                && (FD_ISSET(current->c.fdh.fds, &readset)))
            {
                D_("Calling pluginhandler for fd #%i\n", current->c.fdh.fds);
                (*current->c.fdh.call_module) ();
                D_("pluginhandler for fd #%i returned!\n",
                   current->c.fdh.fds);

                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
            current = current->next;
        }

        /* Now scan through processes */
        tmp = g.active_database;
        while (tmp)
        {
            if ((tmp->start_process)
                && (tmp->start_process->stdout[0] > 0)
                && (FD_ISSET(tmp->start_process->stdout[0], &readset)))
            {
                /* Do the actual read from pipe */
                initng_process_readpipe(tmp, tmp->start_process, FALSE);
                /* If we've been interrupted, return to main loop (program might have ended) */
                if (g.interrupt)
                    return (TRUE);
                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
            if ((tmp->stop_process)
                && (tmp->stop_process->stdout[0] > 0)
                && (FD_ISSET(tmp->stop_process->stdout[0], &readset)))
            {
                /* See above */
                initng_process_readpipe(tmp, tmp->stop_process, FALSE);
                if (g.interrupt)
                    return (TRUE);
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
            tmp = tmp->next;
        }
    }
    return (TRUE);
}
