
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>

#include "initng_global.h"
#include "initng_handler.h"
#include "initng_history.h"
#include "initng_process_db.h"
#include "initng_execute.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "initng_common.h"

#include "main.h"
#include "initng_service_cache.h"
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_fd.h"

#include "initng_kill_handler.h"
#include "initng_static_data_id.h"

static void handle_killed_start(active_h * killed_service);
static void handle_killed_stop(active_h * killed_service);

/* called when a process got killed, identify it, and make a call with a pointer to the process */
void handle_killed_by_pid(pid_t kpid, int r_code)
{
    /* The process that got killed */
    active_h *killd_p = NULL;

    D_("handle_killed_by_pid(%i);\n", kpid);

    /* don't do anything on this pid */
    if (kpid <= 1)
        return;

    /* Look in process database for a match */
    if (!(killd_p = active_db_find_by_pid(kpid)))
    {
        D_("handle_killed_by_pid(%i): No match!\n", kpid);
        return;
    }

    D_("handle_killed_by_pid(%i): found service \"%s\"...\n", kpid,
       killd_p->name);

    if (killd_p->start_process && killd_p->start_process->pid == kpid)
    {
        D_("handle_killed_by_pid(%i): seems to be a start process killed...\n", kpid);
        killd_p->start_process->r_code = r_code;
        handle_killed_start(killd_p);
        return;
    }
    else if (killd_p->stop_process && killd_p->stop_process->pid == kpid)
    {
        D_("handle_killed_by_pid(%i): seems to be a stop process killed...\n",
           kpid);
        killd_p->stop_process->r_code = r_code;
        handle_killed_stop(killd_p);
        return;
    }

    F_("handle_killed_by_pid(%i): traced killed pid to a service but can't find servicetype!\n", kpid);
    return;
}

/* to do when a start action (or a daemon) dies */
static void handle_killed_start(active_h * killed_service)
{
    assert(killed_service);
    assert(killed_service->name);
    g.interrupt = TRUE;

    D_("handle_killed_start(%s);\n", killed_service->name);
    D_(" initial status when called: %s\n",
       active_db_get_status_string(killed_service->a_status));

    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(killed_service, killed_service->start_process,
                            TRUE);

    /* Check if a plugin wants to override handle_killed behavior */
    if (initng_plugin_handle_killed
        (killed_service, killed_service->start_process))
    {
        D_("Plugin did handle this killed.\n");
        return;
    }

    /* print an error message if r_code is positive */
    if (killed_service->start_process->r_code)
        F_("handle_killed_start(%s): Returned with exit %i.\n",
           killed_service->name, killed_service->start_process->r_code);

    /* check so that stop_process is not still running */
    if (killed_service->stop_process)
    {
        W_("Start process stopped, now waiting for stop process to stop!\n");
        return;
    }

    /* if this service has the request to stop, this is good */
    if (killed_service->a_status == STOPPING ||
        killed_service->a_status == STOP_LAUNCHING ||
        killed_service->a_status == STOP_MARKED)
    {
        mark_service(killed_service, STOPPED);
        D_("a_status == aka STOPPING, mark STOPPED and delete from active_db\n");
        goto stop_this;
    }

    /* If this service is a service that is marked up when start
     * process exits, detect this */
    if (active_db_is(&UP_WHEN_DONE, killed_service) ||
        ((active_db_is(initng_sdt_find("start"), killed_service) ||
          active_db_is(initng_sdt_find("start {"), killed_service)) &&
         !active_db_is(&UP_WHEN_STARTED, killed_service)))
    {
        D_("This service, is up when it dies, setting it to up now...\n");
        /* if return code, is positive */
        if (!killed_service->start_process->r_code
            || active_db_is(&UP_WHEN_FAIL, killed_service))
            mark_service(killed_service, DONE);
        else
            mark_service(killed_service, FAIL_STARTING);

        process_db_free(&killed_service->start_process);
        return;
    }

    /* else mark as stopped, check also_stopped, free and forget it */
    mark_service(killed_service, STOPPED);

    /* if this service is set to respawn, do it, but don't respawn services on system state STATE_STOPPING */
    /* or if service is marked restarting, we have to respawn it. */
    if ((active_db_is(&RESPAWN, killed_service))
        && (g.sys_state != STATE_STOPPING))
    {
        D_("Respawn is set, now respawning\n");
        mark_service(killed_service, RESPAWNING);
        process_db_free(&killed_service->start_process);
        /* respawn */
        start_service(killed_service);
        return;
    }

    if (active_db_is_m(killed_service, MARK_RESTARTING))
    {
        D_("Restarting set, now restarting\n");
        mark_service(killed_service, RESTARTING);
        active_db_unmark(killed_service, MARK_RESTARTING);
        process_db_free(&killed_service->start_process);
        /* respawn */
        start_service(killed_service);
        return;
    }

  stop_this:
    /* ok, now aceppt this service as stoping */
    D_("Okay, service %s is really stopping\n", killed_service->name);
    {
        active_h *current = NULL;

        /* mark all services depending on this service stop */
        while ((current = active_db_next_dep_on_me(killed_service, current)))
        {
            D_("%s have to stop %s.\n", killed_service->name, current->name);
            stop_service(current);
        }
    }

    D_("Now removing service %s\n", killed_service->name);
    active_db_del(killed_service);
}


/* to do when a stop action dies */
static void handle_killed_stop(active_h * killed_service)
{
    assert(killed_service);

    g.interrupt = TRUE;
    D_("handle_killed_stop(%s);\n", killed_service->name);

    /* Notify readpipe plugins that we're dead and pipes will be closed */
    initng_process_readpipe(killed_service, killed_service->stop_process,
                            TRUE);

    /* Check if a plugin wants to override handle_killed behavior */
    if (initng_plugin_handle_killed
        (killed_service, killed_service->stop_process))
        return;

    /* free stopped process data */
    process_db_free(&killed_service->stop_process);


    /* mark service stopped */
    mark_service(killed_service, STOPPED);

    if (active_db_is_m(killed_service, MARK_RESTARTING))
    {
        mark_service(killed_service, RESTARTING);
        active_db_unmark(killed_service, MARK_RESTARTING);
        process_db_free(&killed_service->start_process);
        /* respawn */
        start_service(killed_service);
        return;
    }


    /* if there is no start process, count this as the end of service */
    if (killed_service->start_process)
    {
        F_("Bad, there is a start process!\n");
        return;
    }

    /* free service, if stopped = disappear from active list */
    active_db_del(killed_service);
}
