/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dlfcn.h>
#include <assert.h>
#include <string.h>
#include <dirent.h>
#include <fnmatch.h>
#include <errno.h>
#include "initng.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"
#include "initng_main.h"
#include "initng_common.h"

/* 
 * Some modules depend on other modules.
 *
 * A module declares its dependencies by adding an array called
 * module_needs, with the names of the modules that it needs, like
 * this:
 *
 * char *module_needs[] = {
 *     "foo",
 *     "bar",
 *     NULL
 * };
 *
 * When a module is loaded, it checks to see if all of the modules
 * that it depends on are already loaded. If they are not, it refuses
 * to load.
 *
 * The initng_load_all_modules() routine will insure that modules are
 * loaded in the correct order. It does this by loading all modules
 * that have their dependencies met, and iterating until either all
 * modules are loaded or there are no modules left to load. (This is
 * effectively a breadth-first search.)
 *
 * Unloading follows the same logic, except in reverse. A module may
 * not be unloaded if a module that depends on it is loaded. Likewise,
 * the initng_unload_all_modules() routine will unload modules in the
 * proper order. (Again, through a breadth-first search.)
 *
 * Modules that result in circular dependencies can never be loaded.
 *
 * Several things are missing from the depencency setup:
 *
 * - Some notion of "services". A module can depend on only specific
 *   module, say "cpout", not "any module that provides output".
 *
 * - Priority mechanism. The callback functions from modules are
 *   invoked in the order that they are loaded.
 *
 * - Ability for modules to refuse to unload.
 * 
 * Hopefully these will not matter!
 */

/* Note: modules are added to the *start* of the linked list when they
 * are loaded. This makes unloading all modules easy, because a module
 * always sits before the modules that it depends on in the list. */

/* Static includes */
#ifdef STATIC_CHROOT
#include "../plugins/chroot/initng_chroot.h"
#endif


/* 
 * Used to walk through the list of modules one at a time.
 *
 * Returns NULL when there are no modules left.
 */
m_h *initng_get_next_module(m_h * c)
{
    if (c == NULL)
    {
        return g.modules;
    }
    else
    {
        return c->next;
    }
}

/*
 * See if we have already loaded a module with the given name.
 *
 * Returns TRUE if already loaded, else FALSE.
 */
static int module_is_loaded(const char *module_name)
{
    m_h *m;

    assert(module_name != NULL);

    m = g.modules;
    while (m != NULL)
    {
        if (strcmp(m->module_name, module_name) == 0)
        {
            return TRUE;
        }
        m = m->next;
    }
    return FALSE;
}

/* 
 * Returns TRUE if module has its dependencies met, else FALSE.
 */
static int module_needs_are_loaded(const m_h * m)
{
    char **needs;
    int retval;

    assert(m != NULL);

    /* if there are no needs, then we have met them */
    if (m->module_needs == NULL)
    {
        return TRUE;
    }

    /* otherwise check each one */
    needs = m->module_needs;
    retval = TRUE;
    if (needs != NULL)
    {
        while (*needs != NULL)
        {
            if (!module_is_loaded(*needs))
            {
                D_("Module \"%s\" needs \"%s\"\n", m->module_name, *needs);
                retval = FALSE;
            }
            needs++;
        }
    }
    return retval;
}

/* 
 * Read the module information from the file. Does not actually call
 * module_init() for the module, so it is not "loaded" at this point.
 */
int open_module(const char *module_path, const char *module_name, m_h * m)
{
    struct stat st;
    char *errmsg;

    assert(module_path != NULL);
    assert(module_name != NULL);
    assert(m != NULL);

    /* check that file exists */
    if (stat(module_path, &st) != 0)
    {
        F_("Module file \"%s\" not found\n", module_path);
        return FALSE;
    }

    /* open module */
    dlerror();                              /* clear any existing error */
    m->module_dlhandle = dlopen(module_path, RTLD_LAZY);
    /* 
     * this breaks ngc2 on my testbox - neuron : 
     * g.modules[i].module = dlopen(module_name, RTLD_NOW | RTLD_GLOBAL); 
     * */
    if (m->module_dlhandle == NULL)
    {
        F_("Error opening module %s; %s\n", module_name, dlerror());
        return FALSE;
    }

    D_("Success opening module \"%s\"\n", module_name);

    /* get intialisation function */
    dlerror();                              /* clear any existing error */
    m->module_init = dlsym(m->module_dlhandle, "module_init");
    if (m->module_init == NULL)
    {
        errmsg = dlerror();
        dlclose(m->module_dlhandle);
        F_("Error reading module_init(); %s\n", errmsg);
        return FALSE;
    }

    /* get unload function */
    dlerror();                              /* clear any existing error */
    m->module_unload = dlsym(m->module_dlhandle, "module_unload");
    if (m->module_unload == NULL)
    {
        errmsg = dlerror();
        dlclose(m->module_dlhandle);
        F_("Error reading module_unload(); %s\n", errmsg);
        return FALSE;
    }

    /* get dependency list (may be NULL - this is not an error) */
    dlerror();                              /* clear any existing error */
    m->module_needs = dlsym(m->module_dlhandle, "module_needs");
    /* XXX: is there any way to check for "not found", since we don't 
     * consider that an error? */

    /* set module name in database */
    m->module_name = i_strdup(module_name);

    return TRUE;
}

/* 
 * Close the module.
 */
void close_module(m_h * m)
{
    assert(m != NULL);

    dlclose(m->module_dlhandle);
    free(m->module_name);
}

/* load a dynamic module */
/* XXX: more information! */
int initng_load_module(const char *module_name)
{
    char *module_path;
    int ret;
    m_h m;
    m_h *new_m;

    assert(module_name != NULL);

    /* look for duplicates */
    if (module_is_loaded(module_name))
    {
        F_("Module \"%s\" already loaded, won't load it twice\n",
           module_name);
        return (FALSE);
    }

    /* build a path */
    module_path =
        (char *) initng_calloc(1,
                               strlen(INITNG_PLUGIN_DIR) +
                               strlen(module_name) + 9);
    strcpy(module_path, INITNG_PLUGIN_DIR "/lib");
    strcat(module_path, module_name);
    strcat(module_path, ".so\n");

    /* load information from library */
    ret = open_module(module_path, module_name, &m);
    free(module_path);
    if (!ret)
    {
        F_("Unable to load module \"%s\"\n", module_name);
        return (FALSE);
    }

    /* see if we have our dependencies met */
    if (!module_needs_are_loaded(&m))
    {
        F_("Not loading module \"%s\", missing needed module(s)\n",
           module_path);
        close_module(&m);
        return (FALSE);
    }

    ret = (*m.module_init) (INITNG_VERSION);

    D_("for module \"%s\" return: %i\n", module_path, ret);

    if (ret < 1)
    {
        F_("Module %s did not load correctly (module_init() returned %i)\n",
           module_path, ret);
        /* XXX: used to be here, but why? */
        /* sleep(1); */
        close_module(&m);

        return (FALSE);
    }

    /* allocate some memory for the loaded module and add to our list */
    new_m = initng_calloc(1, sizeof(m_h));
    *new_m = m;
    new_m->next = g.modules;
    g.modules = new_m;

    /* and we're done */
    return (TRUE);
}

/*
 * See the named module is needed elsewhere.
 *
 * Returns TRUE if the module is needed, else FALSE.
 */
static int module_is_needed(const char *module_name)
{
    m_h *m;
    char **needs;
    int retval;

    assert(module_name != NULL);

    retval = FALSE;
    m = g.modules;
    while (m != NULL)
    {
        needs = m->module_needs;
        if (needs != NULL)
        {
            while (*needs != NULL)
            {
                if (strcmp(module_name, *needs) == 0)
                {
                    F_("Module \"%s\" needed by \"%s\"\n", module_name,
                       m->module_name);
                    retval = TRUE;
                }
                needs++;
            }
        }
        m = m->next;
    }
    return retval;
}

/*
 * Invoke callback, clean up module.
 */
static void initng_unload_module(m_h * module)
{
    assert(module != NULL);

    (*module->module_unload) ();
    close_module(module);
    free(module);
}

/* XXX: more information! */
int initng_unload_module_named(const char *name)
{
    m_h *prev;
    m_h *this;

    assert(name != NULL);

    D_("initng_load_module_named(%s);\n", name);

    if (!module_is_loaded(name))
    {
        F_("Not unloading module \"%s\", it is not loaded\n", name);
        return FALSE;
    }

    if (module_is_needed(name))
    {
        F_("Not unloading module \"%s\", it is needed\n", name);
        return FALSE;
    }

    /* find the named module in our linked list */
    prev = NULL;
    this = g.modules;
    while (this != NULL)
    {
        if (strcmp(this->module_name, name) == 0)
        {
            break;
        }
        prev = this;
        this = this->next;
    }
    assert(this != NULL);

    /* remove our entry from our list */
    if (this == g.modules)
    {
        g.modules = this->next;
    }
    else
    {
        prev->next = this->next;
    }

    /* and unload the module */
    initng_unload_module(this);
    return TRUE;
}

int initng_load_all_modules(void)
{
    DIR *d;
    struct dirent *e;
    char *module_path;
    int len;
    char module_name[NAME_MAX + 1];

    m_h m;
    m_h *new_m;

    m_h *todo;
    m_h *prev;
    m_h *this;
    int ret;

    /* open plugin dir */
    if (!(d = opendir(INITNG_PLUGIN_DIR)))
    {
        F_("Unable to open plugin directory " INITNG_PLUGIN_DIR ".\n");
        return FALSE;
    }

    /* memory for full path */
    module_path = initng_calloc(strlen(INITNG_PLUGIN_DIR) + NAME_MAX + 2, 1);

    /* get every entry */
    todo = NULL;
    while ((e = readdir(d)) != NULL)
    {
        /* check for files, ending with .so */
        if (fnmatch("lib*.so", e->d_name, 0) == 0)
        {

            /* search the plugin name, for blacklisted */
            if (service_blacklisted(e->d_name))
            {
                F_("Plugin %s blacklisted.\n", e->d_name);
                continue;
            }

            /* set full path */
            strcpy(module_path, INITNG_PLUGIN_DIR "/");
            strcat(module_path, e->d_name);

            /* set name */
            strcpy(module_name, e->d_name + 3); /* skip "lib" */
            len = strlen(module_name);
            module_name[len - 3] = '\0';    /* remove ".so" */

            if (open_module(module_path, module_name, &m))
            {
                /* if we can open it, add it to our "todo" list */
                new_m = initng_calloc(1, sizeof(m_h));
                *new_m = m;
                new_m->next = todo;
                todo = new_m;
            }
        }
        else
        {
#ifdef DEBUG
            if (e->d_name[0] != '.')
            {
                D_("Won't load module \"%s\", doesn't match \"*.so\" pattern.\n", e->d_name);
            }
#endif
        }
    }
    closedir(d);

    free(module_path);

    /* load the entries on our todo list */
    while (todo != NULL)
    {
        /* find the first module that has needs loaded */
        prev = NULL;
        this = todo;
        while (this != NULL)
        {
            if (module_needs_are_loaded(this))
            {
                break;
            }
            prev = this;
            this = this->next;
        }

        /* if we did not find any, then we are done */
        if (this == NULL)
        {
            break;
        }

        /* otherwise remove it from the todo list */
        if (prev == NULL)
        {
            todo = this->next;
        }
        else
        {
            prev->next = this->next;
        }

        /* if we did find find a module with needs loaded, try to load it */
        ret = (*this->module_init) (INITNG_VERSION);
        D_("for module \"%s\" return: %i\n", this->module_name, ret);

        if (ret < 1)
        {
            F_("Module %s did not load correctly (module_init() returned %i)\n", this->module_name, ret);
            close_module(this);
            free(this);
        }
        else
        {
            this->next = g.modules;
            g.modules = this;
        }

    }

    /* if we get to here and there is anything on the todo list, then
     * we have some failed dependencies */
    while (todo != NULL)
    {
        this = todo;
        todo = todo->next;
        F_("Module %s not loaded, needed module(s)\n", this->module_name);
        close_module(this);
        free(this);
    }

    /*    initng_load_static_modules(); */
    return TRUE;
}

#if 0
void initng_load_static_modules(void)
{
    int i;

    /* find a free entry */
    for (i = 0; i < MAX_MODULES; i++)
        if (!g.modules[i].module)
            break;

#ifdef STATIC_CHROOT
    W_("Warning, temporarily setting module_name to libchroot.so.0.0, this should be handled by a struct with name/version that checks name instead of module_name\n");
    strcpy(g.modules[i].module_name, "/lib/initng/libchroot.so.0.0");
    g.modules[i].module = initng_load_static_modules;
    g.modules[i].module_init = chroot_module_init;
    g.modules[i].module_unload = chroot_module_unload;
    i++;
#endif
}
#endif /* 0 */

/* this will dlclose all arrays */
/* XXX: more information needed */

/* 
 * We can simply unload modules in order, because 
 */
void initng_unload_all_modules(void)
{
    m_h *m;

    while (g.modules != NULL)
    {
        m = g.modules;
        g.modules = m->next;
        initng_unload_module(m);
    }

    D_("initng_load_module_close_all()\n");
}
