
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include "initng_process_db.h"
#include "initng_service_cache.h"
#include "initng_global.h"
#include "initng_string_tools.h"
#include "initng_active_db.h"
#include "initng_toolbox.h"
#include "initng_plugin_callers.h"

#include "initng_service_data_types.h"
#include "initng_struct_data.h"

/* add service to data structure, return index or -1 */
service_h *service_db_add(service_h * s)
{
    service_h *current = NULL;

    assert(s);

    /* Check that we are unique */
    if ((current = service_db_find_by_name(s->name)))
    {
        D_("service_db_add(%s): WARNING: DUPLICATE. Already added!\n",
           s->name);
        return (current);
    }

    if (!g.service_cache)
    {
        g.service_cache = s;
        D_("service_db_add(%s): added to cache as first service.\n", s->name);
        return (s);
    }

    current = g.service_cache;
    /* find a free spot */
    while (current)
    {
        if (!current->next)
        {
            current->next = s;
            D_("service_db_add(%s): added to cache next to \"%s\".\n",
               s->name, current->name);
            return (s);
        }

        current = current->next;
    }

    return (NULL);
}



/* returns a service allocated copy from s */
int service_db_copy(service_h ** new_serv, char *name, service_h * s)
{
    /*int i; */

    s_data *our_copy;
    s_data *service_copy;

    assert(s);
    assert(name);
    /* allocate new service */
    *new_serv = (service_h *) initng_calloc(1, sizeof(service_h));
    /* copy service structure */
    memcpy(*new_serv, s, sizeof(service_h));
    /* name has to diff */
    (*new_serv)->name = name;
    /* make a free spot for future data */
    (*new_serv)->data = (s_data *) initng_calloc(1, sizeof(s_data));

    /* walk through all data, and copy all strings. to avoid unwanted surprises. */
    D_("PROPERLY TEST COPY STRINGS!\n");

    service_copy = s->data;
    our_copy = (*new_serv)->data;
    while (service_copy && service_copy->type)
    {
        memcpy(our_copy, service_copy, sizeof(s_data));
        switch (our_copy->type->opt_type)
        {
                /* if type is a string type */
            case STRING:
            case STRINGS:
            case SCRIPT:
                /* copy with i_strdup if string is */
                if (our_copy->t.s)
                {
                    our_copy->t.s = i_strdup(service_copy->t.s);
                    D_("copy string %s\n", our_copy->t.s);
                }
                break;
            default:
                break;
        }

        /* allocate next free spot */
        our_copy->next = (s_data *) initng_calloc(1, sizeof(s_data));

        /* go to next copys */
        our_copy = our_copy->next;
        service_copy = service_copy->next;
    }

    our_copy->next = NULL;

    D_("copy_service(%s);\n", s->name);

    return (TRUE);
}

/* returns an empty service allocated with default values */
int service_db_new(service_h ** d, char *name, e_type type)
{
    assert(name);
    D_("default_service();\n");
    *d = (service_h *) initng_calloc(1, sizeof(service_h));
    if (!*d)
    {
        F_("OUT OF MEMORY default_service().\n");
        return (FALSE);
    }
    (*d)->name = name;
    (*d)->type = type;
    (*d)->data = (s_data *) initng_calloc(1, sizeof(s_data));
    (*d)->data->next = NULL;
    return (TRUE);
}

/* find service by name, returning pointer to service */
service_h *service_db_find_by_name(char *find)
{
    service_h *current = g.service_cache;

    assert(find);

    while (current)
    {
        if (st_service_cmp(find, current->name))
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            return (current);
        }
        current = current->next;
    }
    return (NULL);
}

/* find service by name, returning pointer to service */
service_h *service_db_find_in_name(char *find)
{
    service_h *current = g.service_cache;

    assert(find);

    while (current)
    {
        if (st_service_cmp(find, current->name))
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            return (current);
        }
        current = current->next;
    }

    current = g.service_cache;
    while (current)
    {
        if (strstr(current->name, find))
        {
            D_("service_db_find_by_name(%s): Found\n", current->name);
            return (current);
        }
        current = current->next;
    }


    return (NULL);
}



static int service_db_del(service_h * to_del)
{
    service_h *current = g.service_cache;

    assert(to_del);
    D_("Deleting file-cache %s from struct/chain\n", to_del->name);

    /* check if this is the initial one */
    if (g.service_cache == to_del)
    {
        D_("This is the initial one!\n");
        g.service_cache = (g.service_cache)->next;

        D_("service %s is removed from struct.\n", to_del->name);

        return (TRUE);
    }

    /* walk thru */
    while (current)
    {
        /* check if this is the one */
        if (current->next == to_del)
        {
            current->next = current->next->next;

            D_("service %s is removed from struct.\n", to_del->name);
            return (TRUE);
        }

        /* step forward */
        current = current->next;
    }

    F_("Coud not delete %s from struct.\n", to_del->name);
    return (FALSE);
}

static void clear_father_pointer_from(service_h * s)
{
    service_h *current = g.service_cache;

    while (current)
    {
        if (current->father == s)
            current->father = NULL;

        current = current->next;
    }
}

/* free all dynamically allocated strings in this service */
int service_db_free(service_h * to_free)
{
    assert(to_free);
    D_("service_db_free(%s);\n", to_free->name);

    /* change other database relations to this service to null */
    active_db_change_service_h(to_free, NULL);

    /* this is a dangerous pointer that dont have to hang loose */
    clear_father_pointer_from(to_free);

    /* remove from service db */
    service_db_del(to_free);

    /* free all strings in service */
    if (to_free->data)
    {
        d_remove_all(to_free->data);
        to_free->data = NULL;
    }

    /* free service name */
    if (to_free->name)
        free(to_free->name);
    free(to_free);
    return (TRUE);
}

/* This will do */
void service_db_free_all(void)
{
    while (g.service_cache)
    {
        service_db_free(g.service_cache);
    }
}

int service_db_find_father(service_h * service)
{
    assert(service);
    /* if fahter is known, return ok */
    if (service->father)
        return (TRUE);
    /* if no father is needed */
    if (!service->father_name)
        return (FALSE);

    /* whats for father name in service db */
    service->father = service_db_find_by_name(service->father_name);
    if (service->father)
        return (TRUE);

    /* try to load father */
    service->father = initng_modules_parse(service->father_name);
    if (service->father)
        return (TRUE);

    /* else fail */
    return (FALSE);
}

char *service_db_get_string(s_entry * type, service_h * s)
{
    char *tmp = NULL;

    assert(type);
    assert(s);

    /* try to find in local data */
    if ((tmp = d_get_string(type, s->data)))
        return (tmp);

    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if ((tmp = service_db_get_string(type, s->father)))
            return (tmp);

    return (NULL);
}

char **service_db_get_stringlist(s_entry * type, service_h * s)
{
    char **tmp = NULL;

    assert(type);
    assert(s);

    /* try to find in local data */
    if ((tmp = d_get_stringlist(type, s->data)))
        return (tmp);

    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if ((tmp = service_db_get_stringlist(type, s->father)))
            return (tmp);

    return (NULL);
}

int service_db_get_int(s_entry * type, service_h * s)
{
    assert(type);
    assert(s);
    /* try to find in local data */
    if (d_is(type, s->data))
        return (d_get_int(type, s->data));
    /* look for data in my parrent, this is bad and wont look in parents parent. */
    if (service_db_find_father(s) && s->father && d_is(type, s->father->data))
        return (d_get_int(type, s->data));
    return (0);
}

char *service_db_get_next_string(s_entry * type, service_h * s, char *last)
{
    char *tmp = NULL;

    assert(type);
    assert(s);
    /* try to find in local data */
    if ((tmp = d_get_next_string(type, s->data, last)))
        return (tmp);
    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if ((tmp = service_db_get_next_string(type, s->father, last)))
            return (tmp);
    return (NULL);
}

int service_db_is(s_entry * type, service_h * s)
{
    assert(type);
    assert(s);
    if (d_is(type, s->data))
        return (TRUE);
    /* look for data in my parrent */
    if (service_db_find_father(s) && s->father)
        if (service_db_is(type, s->father))
            return (TRUE);
    return (FALSE);
}
